//bridge-file-version: #184


import { ChestFormData } from './extensions/forms.js';
import { ActionFormData, MessageFormData, ModalFormData } from "@minecraft/server-ui";


import { world, system, Player, Entity, ItemStack, EntityDamageCause, EquipmentSlot, ScoreboardObjective } from '@minecraft/server';
import { VECTOR3_ZERO, Vector3Utils } from './utils/coreHelpers';
import { getObjective, getScore } from './utils/scoreUtils';
import { config } from './config';
import './settings';
import './index';
//import './interact';
/**
 * アイテムの種類を返す
 * @param {ItemStack} itemStack 
 * @returns {String}
 */
function getReplaceTypeId(itemStack) {
    return (itemStack?.typeId || 'air').replace('minecraft:', '');
}

/**
 * タグ付与
 * @param {Entity} entity 
 * @param {String[]} tags 
 */
function addTags(entity, tags) {
    if (tags.length > 0)
    for (const player of world.getPlayers({ tags: ['status:debug'] })) {
        player.sendMessage(`[${entity?.nameTag || entity.typeId.replace('minecraft:', '')}] ${tags.join(' | ')}`);
    }
    tags.forEach(tag => entity.addTag(tag));
}

/**
 * タグ削除
 * @param {Entity} entity 
 * @param {String[]} tags 
 */
function removeTags(entity, tags) {
    tags.forEach(tag => entity.removeTag(tag));
}

/**
 * タグ付与と削除をセットで行う
 * @param {Entity} entity 
 * @param {String[]} tags 
 * @param {Number} timeout 
 */
function addAndRemoveTags(entity, tags, timeout = 1) {
    if (tags.length > 0) {
        addTags(entity, tags);
        system.runTimeout(() => {
            if (entity.isValid()) removeTags(entity, tags);
        }, timeout);
    }
}

const useItemMap = new Map();
const arrowShotMap = new Map();
const arrowDataMap = new Map();

// ブロック破壊イベント
world.beforeEvents.playerBreakBlock.subscribe(ev => {
    const pb = config.behavior.settings.playerBreakBlock.settings;
    if (!pb.all.enabled) return;

    const { player, block, itemStack } = ev;
    const { x, y, z } = block.location;
    const tags = [
        pb.break.enabled && 't:break',
        pb.itemid.enabled && 't:break_itemid_' + getReplaceTypeId(itemStack),
        pb.itemname.enabled && itemStack?.nameTag && 't:break_itemname_' + itemStack.nameTag,
        pb.location.enabled && 't:break_location_' + [x, y, z].join(','),
    ].filter(Boolean);

    system.run(() => {
        addAndRemoveTags(player, tags);
    });
});

world.beforeEvents.itemUseOn.subscribe(ev => {
    const { source, itemStack } = ev;
    useItemMap.set(source.id, itemStack.typeId);
    system.runTimeout(() => {
        useItemMap.delete(source.id);
    }, 1);
});

// ブロック設置イベント
world.afterEvents.playerPlaceBlock.subscribe(ev => {
    const pp = config.behavior.settings.playerPlaceBlock.settings;
    if (!pp.all.enabled) return;

    const { player, block } = ev;
    const { x, y, z } = block.location;
    const item = useItemMap.get(player.id);
    const blockId = item ? item.replace('minecraft:', '') : '';

    const tags = [
        pp.place.enabled && 't:place',
        pp.block.enabled && 't:place_block_' + blockId,
        pp.location.enabled && 't:place_location_' + [x, y, z].join(',')
    ].filter(Boolean);

    addAndRemoveTags(player, tags);
});

// 投射物がブロックに当たったイベント
world.afterEvents.projectileHitBlock.subscribe(ev => {
    const bu = config.behavior.settings.bowUse.settings;
    const ar = config.behavior.settings.arrow.settings;
    const { projectile, source } = ev;

    if (projectile.typeId === 'minecraft:arrow' && source?.isValid()) {
        if (bu.hitBlock.enabled) addAndRemoveTags(source, ['t:hit_block']);
        if (ar.hitBlock.enabled) addAndRemoveTags(projectile, ['t:arrow_hit_block']);

        if (source.hasTag('arrow_remove_on_hit')) {
            if (projectile.isValid()) projectile.kill();
        }
        else if (source.hasTag('arrow_remove_on_hit_2')) {
            system.runTimeout(() => {
                if (projectile.isValid()) projectile.kill();
            }, 5)
        }
    }
});

// アイテム使用イベント
world.afterEvents.itemUse.subscribe(ev => {
    const iu = config.behavior.settings.itemUse.settings;
    const bu = config.behavior.settings.bowUse.settings;
    const ar = config.behavior.settings.arrow.settings;

    const { source, itemStack } = ev;
    const tags = [
        iu.all.enabled && iu.use.enabled && 't:use',
        iu.all.enabled && iu.itemid.enabled && 't:use_itemid_' + getReplaceTypeId(itemStack),
        iu.all.enabled && iu.itemname.enabled && itemStack.nameTag && 't:use_itemname_' + itemStack.nameTag,
    ];

    const arrowId = arrowShotMap.get(source.id);
    if (itemStack.typeId === 'minecraft:crossbow' && arrowId) {
        arrowShotMap.delete(source.id);
        arrowDataMap.set(arrowId, { typeId: itemStack.typeId, nameTag: itemStack.nameTag });

        tags.push(
            bu.all.enabled && bu.crossbowShot.enabled && 't:crossbow_shot',
            bu.all.enabled && bu.itemname.enabled && itemStack.nameTag && 't:crossbow_shot_itemname_' + itemStack.nameTag,
        );

        const arrow = source.dimension.getEntities({ type: 'minecraft:arrow' }).find(e => e.id === arrowId);
        if (arrow) {
            const arrowTags = [
                ar.crossbow.enabled && 't:arrow_crossbow',
                ar.itemname.enabled && itemStack.nameTag && 't:arrow_itemname_' + itemStack.nameTag,
            ].filter(Boolean);

            addTags(arrow, arrowTags);
        }
    }

    addAndRemoveTags(source, tags.filter(Boolean));
});


// アイテム使用解除イベント
world.afterEvents.itemReleaseUse.subscribe(ev => {
    const ar = config.behavior.settings.arrow.settings;
    const bu = config.behavior.settings.bowUse.settings;
    const { itemStack, source } = ev;
    const arrowId = arrowShotMap.get(source.id);

    if (itemStack.typeId === 'minecraft:bow' && arrowId) {
        const arrow = source.dimension.getEntities({ type: 'minecraft:arrow' }).find(e => e.id === arrowId);
        if (arrow) {
            arrowShotMap.delete(source.id);
            arrowDataMap.set(arrowId, { typeId: itemStack.typeId, nameTag: itemStack?.nameTag });
            const tags = [
                bu.all.enabled && bu.bowShot.enabled && 't:bow_shot',
                bu.all.enabled && bu.itemname.enabled && itemStack?.nameTag && 't:bow_shot_itemname_' + itemStack?.nameTag,
            ].filter(Boolean);
            const arrowTags = [
                ar.bow.enabled && 't:arrow_bow',
                ar.itemname.enabled && itemStack?.nameTag && 't:arrow_itemname_' + itemStack?.nameTag,
            ].filter(Boolean);

            addAndRemoveTags(source, tags);
            addTags(arrow, arrowTags);
        }
    }
});

// アイテム使用完了イベント
world.afterEvents.itemCompleteUse.subscribe(ev => {
    const icu = config.behavior.settings.itemCompleteUse.settings;
    if (!icu.all.enabled) return;

    const { source, itemStack } = ev;
    const tags = [
        icu.complateUse.enabled && 't:complet_use',
        icu.complateUse.enabled && 't:complet_use_itemid_' + getReplaceTypeId(itemStack),
        icu.itemname.enabled && itemStack?.nameTag && 't:complet_use_itemname_' + itemStack?.nameTag,
    ].filter(Boolean);

    addAndRemoveTags(source, tags);
});

// プレイヤーがスポーンしたイベント
world.afterEvents.playerSpawn.subscribe(ev => {
    const oh = config.behavior.settings.other.settings;
    const { player, initialSpawn } = ev;
    const tags = oh.initialSpawn.enabled && initialSpawn ? ['t:initial_spawn'] : oh.respawn.enabled ? ['t:respawn'] : [];
    addAndRemoveTags(player, tags);
});

// エンティティがスポーンしたイベント
world.afterEvents.entitySpawn.subscribe(ev => {
    const { entity, cause } = ev;
    if (entity.typeId === 'minecraft:arrow') {
        const player = entity.getComponent('projectile')?.owner;
        if (player?.typeId === 'minecraft:player') {
            arrowShotMap.set(player.id, entity.id);
            system.runTimeout(() => {
                arrowShotMap.delete(player.id);
            }, 1);
        }
    }
});

// エンティティのヘルスが変更されたイベント
world.afterEvents.entityHealthChanged.subscribe(ev => {
    if (!config.recordingScore.settings.heal.enabled || ev.entity.typeId !== 'minecraft:player') return;
    const { entity, oldValue, newValue } = ev;
    const healValue = newValue - oldValue;
    const heal = getObjective('heal');

    if (healValue > 0 && heal) {
        heal.addScore(entity, healValue);
    }
});

// エンティティがダメージを受けたイベント
world.afterEvents.entityHurt.subscribe(ev => {
    const eh = config.behavior.settings.entityHurt.settings;
    const rs = config.recordingScore.settings;

    const { hurtEntity, damageSource: { damagingEntity, damagingProjectile, cause }, damage } = ev;

    const damageTakenCount = getObjective('damageTakenCount');
    const damageTaken = getObjective('damageTaken');
    const damageDealtCount = getObjective('damageDealtCount');
    const damageDealt = getObjective('damageDealt');
    const currentDamageDealt = getObjective('currentDamageDealt');
    const currentDamageTaken = getObjective('currentDamageTaken');

    const damagedTags = [
        eh.all.enabled && eh.damaged.enabled && 't:damaged',
        eh.all.enabled && eh.cause.enabled && `t:d_cause_${cause}`,
        eh.all.enabled && eh.entity.enabled && damagingEntity && `t:d_entity`,
        eh.all.enabled && eh.entityId.enabled && damagingEntity && `t:d_entity_${damagingEntity?.typeId?.replace('minecraft:', '')}`,
        eh.all.enabled && eh.projectile.enabled && damagingProjectile && `t:d_projectile_${damagingProjectile.typeId.replace('minecraft:', '')}`,
    ];

    const attackingTags = [
        eh.all.enabled && eh.attack.enabled && 't:attack',
        eh.all.enabled && eh.cause.enabled && `t:a_cause_${cause}`,
        eh.all.enabled && eh.entity.enabled && `t:a_entity`,
        eh.all.enabled && eh.entityId.enabled && `t:a_entity_${hurtEntity?.typeId?.replace('minecraft:', '')}`,
        eh.all.enabled && eh.projectile.enabled && damagingProjectile && `t:a_projectile_${damagingProjectile.typeId.replace('minecraft:', '')}`,
    ];

    if (damagingEntity) {
        if (cause === EntityDamageCause.entityAttack || damagingProjectile?.typeId === 'minecraft:arrow') {
            let weaponItem = damagingEntity?.getComponent('equippable')?.getEquipment(EquipmentSlot.Mainhand) || { typeId: 'air' };

            if (damagingProjectile?.typeId === 'minecraft:arrow') {
                const arrowData = arrowDataMap.get(damagingProjectile.id);
                if (arrowData) {
                    weaponItem = arrowData;
                }
            }
            if (eh.all.enabled && eh.itemid.enabled) {
                damagedTags.push(`t:d_itemid_${getReplaceTypeId(weaponItem)}`);
                attackingTags.push(`t:a_itemid_${getReplaceTypeId(weaponItem)}`);
            }
            if (eh.all.enabled && eh.itemname.enabled && weaponItem?.nameTag) {
                damagedTags.push(`t:d_itemname_${weaponItem.nameTag}`);
                attackingTags.push(`t:a_itemname_${weaponItem.nameTag}`);
            }
        }
    }

    addAndRemoveTags(hurtEntity, damagedTags.filter(Boolean));
    if (rs.damageTakenCount.enabled && damageTakenCount) damageTakenCount.addScore(hurtEntity, 1);
    if (rs.damageTaken.enabled && damageTaken) damageTaken.addScore(hurtEntity, damage);
    if (rs.currentDamageTaken.enabled && currentDamageTaken) currentDamageTaken.addScore(hurtEntity, damage);

    if (damagingEntity?.typeId === 'minecraft:player') {
        addAndRemoveTags(damagingEntity, attackingTags.filter(Boolean));

        if (rs.damageDealtCount.enabled && damageDealtCount) damageDealtCount.addScore(damagingEntity, 1);
        if (rs.damageDealt.enabled && damageDealt) damageDealt.addScore(damagingEntity, damage);
        if (rs.currentDamageDealt.enabled && currentDamageDealt) currentDamageDealt.addScore(damagingEntity, damage);

        if (cause === EntityDamageCause.projectile && damagingEntity.hasTag('sound_hit')) {
            damagingEntity.playSound('random.orb', { volume: 0.5 });
        }
    }

    system.runTimeout(() => {
        if (rs.currentDamageTaken.enabled && currentDamageTaken) currentDamageTaken.setScore(hurtEntity, 0);
        if (rs.currentDamageDealt.enabled && damagingEntity && currentDamageDealt) currentDamageDealt.setScore(damagingEntity, 0);
    }, 1);
});


// エンティティが死亡したイベント
world.afterEvents.entityDie.subscribe(ev => {
    const ed = config.behavior.settings.entityDie.settings;
    const rs = config.recordingScore.settings;

    const { damageSource: { damagingEntity }, deadEntity } = ev;
    const killCount = getObjective('killCount');
    const deathCount = getObjective('deathCount');

    const deathTags = [
        ed.all.enabled && ed.death.enabled && 't:death',
    ].filter(Boolean);
    const killTags = [
        ed.all.enabled && ed.kill.enabled && 't:kill',
    ].filter(Boolean);

    if (deadEntity?.typeId === 'minecraft:player') {
        addAndRemoveTags(deadEntity, deathTags);
        if (rs.deathCount.enabled && deathCount) deathCount.addScore(deadEntity.scoreboardIdentity, 1);
    }
    if (damagingEntity?.typeId === 'minecraft:player' && deadEntity?.typeId === 'minecraft:player') {
        addAndRemoveTags(damagingEntity, killTags);
        if (rs.killCount.enabled && killCount) killCount.addScore(damagingEntity.scoreboardIdentity, 1);
    }
});

system.runInterval(() => {
    updateStatus();
});

function updateStatus() {
    const objectives = {
        'jumpCount': getObjective('jumpCount'),
        'speed_m_s': getObjective('speed_m_s'),
        'speed_km_h': getObjective('speed_km_h'),
        'distance': getObjective('distance'),
        'distance_m': getObjective('distance_m'),
        'hp': getObjective('HP'),
        'scoreX': getObjective('x'),
        'scoreY': getObjective('y'),
        'scoreZ': getObjective('z'),
        'scoreRx': getObjective('rx'),
        'scoreRy': getObjective('ry'),
    }

    world.getPlayers().forEach(player => new PlayerStatus(player, objectives));
}

const eventTags = [
    'can_not_climb',
    'stackable',
    'noname',
    'player_ride'
];

const statusObjectives = [
    ['health', 20, 0, 300, 'health'],
    ['movement', 10, 0, 300, 'movement'],
    ['attack', 1, 0, 300, 'attack'],
    ['scale', 1, 1, 300, 'scale'],
    ['knockback_resi', 0, -300, 300, 'knockback_resistance'],
    ['arrow_damage', 1, 0, 300, 'status:arrow_damage'],
];

class PlayerStatus {
    /**
     * 
     * @param {Player} player 
     * @param {Object.<string, ScoreboardObjective>} objectives 
     */
    constructor(player, objectives) {
        this.player = player;
        this.objectives = objectives;
        this.tags = player.getTags();
        this.velocity = player.getVelocity();
        this.location = player.location;
        this.rotation = player.getRotation();
        this.equippable = player.getComponent('equippable');
        this.mainhand = this.equippable?.getEquipment(EquipmentSlot.Mainhand);
        this.offhand = this.equippable?.getEquipment(EquipmentSlot.Offhand);

        this.reset();
        this.eventTags();
        this.updateScores();
        this.updateTags();
        this.equipmentTags();
        this.setPlayerStatus();

        if (player.hasTag('status:debug')) {
            player.onScreenDisplay.setActionBar(player.getTags().join("\n"));
        }
    }

    reset() {
        this.player.triggerEvent('reset');
    }

    eventTags() {
        eventTags.forEach(tag => {
            if (this.tags.includes(tag)) {
                this.player.triggerEvent(tag);
            }
        });
    }

    updateScores() {
        const { jumpCount, distance, distance_m, speed_m_s, speed_km_h, hp, scoreX, scoreY, scoreZ, scoreRx, scoreRy } = this.objectives;
        const player = this.player;
        const pb = config.behavior.settings.playerBehavior.settings;
        const rs = config.recordingScore.settings;

        if (rs.hp.enabled && hp) {
            const health = player.getComponent('health').currentValue;
            hp.setScore(player, Math.max(health, 0));
        }

        if (player.isJumping && !player.isOnGround && player.hasTag('t:on_ground')) {
            if (pb.jumping.enabled) {
                player.addTag('t:jumping');
                system.runTimeout(() => { player.removeTag('t:jumping') }, 1);
            }
            if (rs.jumpCount.enabled && jumpCount) jumpCount.addScore(player, 1);
        }

        const distanceValue = Vector3Utils.distance(VECTOR3_ZERO, { x: this.velocity.x, y: 0, z: this.velocity.z });
        if (rs.distance.enabled && distance) {
            distance.addScore(player, Math.floor(distanceValue * 1000));
            if (distance_m) distance_m.setScore(player, distance.getScore(player) / 1000);
        }
        if (rs.speed_m_s.enabled && speed_m_s) speed_m_s.setScore(player, Math.floor(distanceValue * 20));
        if (rs.speed_km_h.enabled && speed_km_h) speed_km_h.setScore(player, Math.floor(distanceValue * 20 * 60 * 60 / 1000));

        const { x, y, z } = this.location;
        const { x: rx, y: ry } = this.rotation;
        if (rs.x.enabled && scoreX) scoreX.setScore(player, Math.floor(x));
        if (rs.y.enabled && scoreY) scoreY.setScore(player, Math.floor(y));
        if (rs.z.enabled && scoreZ) scoreZ.setScore(player, Math.floor(z));

        if (rs.rx.enabled && scoreRx) scoreRx.setScore(player, rx);
        if (rs.ry.enabled && scoreRy) scoreRy.setScore(player, ry);
    }

    updateTags() {
        const pb = config.behavior.settings.playerBehavior.settings;
        const oh = config.behavior.settings.other.settings;
        const po = config.behavior.settings.position.settings;

        const conditions = [
            [oh.alive.enabled && this.player.getComponent('health').currentValue > 0, 't:alive'],
            [oh.sleeping.enabled && this.player.isSleeping, 't:sleeping'],
            [pb.all.enabled && pb.sneaking.enabled && this.player.isSneaking, 't:sneaking'],
            [pb.all.enabled && pb.jumpingAlways.enabled && this.player.isJumping, 't:jumping_always'],
            [pb.all.enabled && pb.guarding.enabled && this.player.isSneaking && (this.mainhand?.typeId === 'minecraft:shield' || this.offhand?.typeId === 'minecraft:shield'), 't:guarding'],
            [pb.all.enabled && pb.sprinting.enabled && this.player.isSprinting, 't:sprinting'],
            [pb.all.enabled && pb.swimming.enabled && this.player.isSwimming, 't:swimming'],
            [pb.all.enabled && pb.climbing.enabled && this.player.isClimbing, 't:climbing'],
            [pb.all.enabled && pb.flying.enabled && this.player.isFlying, 't:flying'],
            [pb.all.enabled && pb.glinding.enabled && this.player.isGliding, 't:gliding'],
            [pb.all.enabled && pb.falling.enabled && this.player.isFalling, 't:falling'],
            [pb.all.enabled && pb.moving.enabled && Vector3Utils.magnitude({ x: this.velocity.x, y: this.velocity.y, z: this.velocity.z }) > 0.003, 't:moving'],
            [po.onGround.enabled && this.player.isOnGround, 't:on_ground'],
            [po.inWater.enabled && this.player.isInWater, 't:in_water'],
        ];
        conditions.forEach(([condition, tag]) => {
            if (condition) {
                this.player.addTag(tag);
            } else {
                this.player.removeTag(tag);
            }
        });
    }

    equipmentTags() {
        const hi = config.behavior.settings.hasItem.settings;
        const di = config.behavior.settings.position.settings.dimension.enabled;
        const equipmentTags = [
            ['t:mainhand', hi.all.enabled && hi.mainhand.enabled && this.mainhand],
            ['t:offhand', hi.all.enabled && hi.offhand.enabled && this.offhand],
            ['t:head', hi.all.enabled && hi.head.enabled && this.equippable.getEquipment(EquipmentSlot.Head)],
            ['t:chest', hi.all.enabled && hi.chest.enabled && this.equippable.getEquipment(EquipmentSlot.Chest)],
            ['t:legs', hi.all.enabled && hi.legs.enabled && this.equippable.getEquipment(EquipmentSlot.Legs)],
            ['t:feet', hi.all.enabled && hi.feet.enabled && this.equippable.getEquipment(EquipmentSlot.Feet)],
        ];

        equipmentTags.forEach(([tag, item]) => {
            const existingTags = this.tags.filter(t => t.startsWith(tag));
            existingTags.forEach(existingTag => {
                this.player.removeTag(existingTag);
            });

            if (item) {
                if (hi.itemid.enabled) this.player.addTag(`${tag}_itemid_${item.typeId.replace('minecraft:', '')}`);
                if (hi.itemname.enabled && item.nameTag) this.player.addTag(`${tag}_itemname_${item.nameTag}`);
            }
        });

        const dimensionTag = this.tags.find(t => t.startsWith('t:dimension_'));
        if (dimensionTag) this.player.removeTag(dimensionTag);
        if (di) {
            this.player.addTag('t:dimension_' + this.player.dimension.id.replace('minecraft:', ''));
        }
    }

    setPlayerStatus() {
        statusObjectives.forEach(([objectiveId, defaultScore, min, max, event]) => {
            const score = getScore(this.player, objectiveId);
            if (objectiveId === 'arrow_damage') {
                this.player.setProperty(event, score === undefined || !(score >= min && score <= max) ? -1 : score);
            } else {
                if (score === undefined || !(score >= min && score <= max)) {
                    this.player.triggerEvent(event + '_reset');
                } else {
                    this.player.triggerEvent(event + score);
                }
            }
        });
    }
}

